/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Parsing hooks for color output in terminals
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>

#include "parse.h"


struct gen_cmds_context {
	FILE *out;            /* output stream */

	unsigned int indent;  /* indentation size */
};



/* --------------------------------------------------------------------------
 *   color and font definitions
 */

/* normal */
#define COLOR_N_RED      "\033[00;31m"
#define COLOR_N_GREEN    "\033[00;32m"
#define COLOR_N_YELLOW   "\033[00;33m"
#define COLOR_N_BLUE     "\033[00;34m"
#define COLOR_N_MAGENTA  "\033[00;35m"
#define COLOR_N_CYAN     "\033[00;36m"

/* bold */
#define COLOR_B_RED      "\033[01;31m"
#define COLOR_B_GREEN    "\033[01;32m"
#define COLOR_B_YELLOW   "\033[01;33m"
#define COLOR_B_BLUE     "\033[01;34m"
#define COLOR_B_MAGENTA  "\033[01;35m"
#define COLOR_B_CYAN     "\033[01;36m"

/* dim */
#define COLOR_D_BLUE     "\033[02;34m"
#define COLOR_D_CYAN     "\033[02;36m"

/* coloring statements */
#define COLOR_COMMAND       COLOR_B_BLUE
#define COLOR_COMMAND_META  COLOR_D_CYAN
#define COLOR_OUTPUT        COLOR_N_RED
#define COLOR_COMMENT       COLOR_N_GREEN

/* back to normal */
#define COLOR_RESET      "\033[00;39m"



/* --------------------------------------------------------------------------
 *   parsing operations
 */

static void *
init(FILE *out)
{
	struct gen_cmds_context *ctx;

	ctx = malloc(sizeof(*ctx));
	if (!ctx)
		return NULL;

	ctx->out = out;

	return ctx;
}

static void
cleanup(void *priv)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;

	fclose(ctx->out);
	free(ctx);
}

static void
set_indent(void *priv, unsigned int width)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;

	ctx->indent = width;
}

static int
line_command(void *priv,
             parse_state prev_state,
             char *line,
             char *data,
             size_t len)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;
	size_t last = len - 1;

	if (data[last] == '\n')
		last--;

	switch (prev_state) {
	case STATE_COMMAND:
	case STATE_OUTPUT:
	case STATE_NONE:
		fputs(COLOR_COMMAND_META, ctx->out);
		fwrite(line, ctx->indent + 1, 1, ctx->out);
		fputs(COLOR_COMMAND, ctx->out);
		if (data[last] == '\\') {
			fwrite(line + ctx->indent + 1, last + 1, 1, ctx->out);
			fputs(COLOR_COMMAND_META "\\\n", ctx->out);
		} else {
			fputs(line + ctx->indent + 1, ctx->out);
		}
		break;
	default:
		assert(0);
		return -1;
	}

	return 0;
}

static int
line_continuation(void *priv,
                  parse_state prev_state,
                  char *line,
                  char *data,
                  size_t len)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;
	size_t last = len - 1;

	if (data[last] == '\n')
		last--;

	switch (prev_state) {
	case STATE_COMMAND:
		fputs(COLOR_COMMAND_META, ctx->out);
		fwrite(line, ctx->indent + 1, 1, ctx->out);
		fputs(COLOR_COMMAND, ctx->out);
		if (data[last] == '\\') {
			fwrite(line + ctx->indent + 1, last + 1, 1, ctx->out);
			fputs(COLOR_COMMAND_META "\\\n", ctx->out);
		} else {
			fputs(line + ctx->indent + 1, ctx->out);
		}
		return 0;
	case STATE_OUTPUT:
	case STATE_NONE:
	default:
		assert(0);
		return -1;
	}
}

static int
line_output(void *priv,
            parse_state prev_state,
            char *line,
            char *data,
            size_t len)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;

	switch (prev_state) {
	case STATE_COMMAND:
	case STATE_OUTPUT:
		fputs(COLOR_COMMAND_META, ctx->out);
		fwrite(line, ctx->indent, 1, ctx->out);
		fputs(COLOR_OUTPUT, ctx->out);
		fputs(data, ctx->out);
		return 0;
	case STATE_NONE:
	default:
		assert(0);
		return -1;
	}
}

static int
line_other(void *priv,
           parse_state prev_state,
           char *line,
           char *data,
           size_t len)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;

	switch (prev_state) {
	case STATE_COMMAND:
	case STATE_OUTPUT:
	case STATE_NONE:
		fputs(COLOR_COMMENT, ctx->out);
		fputs(data, ctx->out);
		return 0;
	default:
		assert(0);
		return -1;
	}
}

static int
finalize(void *priv, parse_state prev_state)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;

	fputs(COLOR_RESET, ctx->out);

	return 0;
}



/* --------------------------------------------------------------------------
 *   operations table
 */

const struct parse_ops parse_ops_gen_ansi = {
	.init               = init,
	.set_indent         = set_indent,
	.line_command       = line_command,
	.line_continuation  = line_continuation,
	.line_output        = line_output,
	.line_other         = line_other,
	.processed_line     = NULL,
	.finalize           = finalize,
	.cleanup            = cleanup,
};
