/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Parsing hooks for generating shell commands
 *
 * These commands, when executed by a bourne shell, will generate an
 * output suitable for reading with an istream.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>

#include "parse.h"
#include "wbuf.h"


struct gen_cmds_context {
	FILE *out;               /* output stream */

	char *salt;              /* random salt string */
	unsigned int indent;     /* indentation size */

	unsigned int keep;       /* number of command and comment lines */
	unsigned int skip;       /* number of command output lines */

	parse_state state;       /* current parsing state */
	parse_state prev_state;  /* previous parsing state */

	struct wbuf *cmd;        /* command lines buffer */
};



/* --------------------------------------------------------------------------
 *   private functions
 */

static char *
generate_salt(char *buf, size_t len)
{
	FILE *fp;
	char *p;

	p = getenv("KUULTO_SALT");
	if (p && *p) /* environment override */
		return p;

	fp = fopen("/dev/urandom", "r");
	if (!fp) {
		snprintf(buf, len, "SALT%d%d", getpid(), getppid());
		return buf;
	}
	p = buf;
	while (--len)
		*p++ = ((fgetc(fp) % 2) ? 'a' : 'A') + (fgetc(fp) % 26);
	*p = 0;
	fclose(fp);

	return buf;
}

static void
emit_salt(struct gen_cmds_context *ctx,
          parse_state prev_state,
          parse_state state)
{
	static const char * const states = "*=#.";

	ctx->keep -= ctx->skip;

	fprintf(ctx->out, "echo \"%s $? %d %d %c%c @\"\n", ctx->salt,
	                                                   ctx->keep,
	                                                   ctx->skip,
	                                                   states[prev_state],
	                                                   states[state]);

	ctx->keep = 0;
	ctx->skip = 0;
}

static void
write_header(struct gen_cmds_context *ctx)
{
	fprintf(ctx->out, "echo \"%s %s %u\"\n", "|klp|", ctx->salt, ctx->indent);
}

static int
command(void *priv, char *data)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;

	return (fputs(data, ctx->out) == EOF) ? -1 : 0;
}



/* --------------------------------------------------------------------------
 *   parsing operations
 */

static void *
init(FILE *out)
{
	struct gen_cmds_context *ctx;
	char saltbuf[12], *salt;
	struct wbuf *cmd;

	if ( !(ctx = malloc(sizeof(*ctx))) ||
	     !(cmd  = wbuf_init())         ||
	     !(salt = strdup(generate_salt(saltbuf, sizeof(saltbuf)))) )
	{
		if (ctx)
			free(cmd);
		free(ctx);
		return NULL;
	}

	ctx->out         = out;
	ctx->salt        = salt;
	ctx->indent      = 0;
	ctx->keep        = 0;
	ctx->skip        = 0;
	ctx->state       = STATE_NONE;
	ctx->prev_state  = STATE_NONE;
	ctx->cmd         = cmd;

	return ctx;
}

static void
cleanup(void *priv)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;

	fclose(ctx->out); /* send eof to consumer */
	wbuf_free(ctx->cmd);
	free(ctx->salt);
	free(ctx);
}

static void
processed_line(void *priv)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;

	ctx->keep++;
}

static int
finalize(void *priv, parse_state prev_state)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;
	int res = 0;

	if (!ctx->indent)
		/* no command found, but there might be content still */
		write_header(ctx);

	if (!wbuf_empty(ctx->cmd)) {
		emit_salt(ctx, prev_state, STATE_END);
		res = (wbuf_yield(ctx->cmd, command, ctx) == -1) ? -1 : 0;
	}

	emit_salt(ctx, STATE_END, STATE_END);

	return res;
}

static void
set_indent(void *priv, unsigned int width)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;

	ctx->indent = width;
	write_header(ctx);
}

static int
line_command(void *priv,
             parse_state prev_state,
             char *line,
             char *data,
             size_t len)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;

	switch (prev_state) {
	case STATE_COMMAND:
	case STATE_OUTPUT:
		emit_salt(ctx, prev_state, STATE_COMMAND);
		if (wbuf_yield(ctx->cmd, command, ctx) == -1)
			return -1;
		/* fall through */
	case STATE_NONE:
		return wbuf_append(ctx->cmd, data);
	default:
		assert(0);
		return -1;
	}
}

static int
line_continuation(void *priv,
                  parse_state prev_state,
                  char *line,
                  char *data,
                  size_t len)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;

	switch (prev_state) {
	case STATE_COMMAND:
		return wbuf_append(ctx->cmd, data);
	case STATE_OUTPUT:
		emit_salt(ctx, prev_state, STATE_NONE);
		return wbuf_yield(ctx->cmd, command, ctx);
	case STATE_NONE:
		return 0;
	default:
		assert(0);
		return -1;
	}
}

static int
line_output(void *priv,
            parse_state prev_state,
            char *line,
            char *data,
            size_t len)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;

	switch (prev_state) {
	case STATE_COMMAND:
	case STATE_OUTPUT:
		ctx->skip++;
		/* fall through */
	case STATE_NONE:
		return 0;
	default:
		assert(0);
		return -1;
	}
}

static int
line_other(void *priv,
           parse_state prev_state,
           char *line,
           char *data,
           size_t len)
{
	struct gen_cmds_context *ctx = (struct gen_cmds_context *)priv;

	switch (prev_state) {
	case STATE_COMMAND:
	case STATE_OUTPUT:
		emit_salt(ctx, prev_state, STATE_NONE);
		return wbuf_yield(ctx->cmd, command, ctx);
	case STATE_NONE:
		return 0;
	default:
		assert(0);
		return -1;
	}
}



/* --------------------------------------------------------------------------
 *   operations table
 */

const struct parse_ops parse_ops_gen_cmds = {
	.init               = init,
	.set_indent         = set_indent,
	.line_command       = line_command,
	.line_continuation  = line_continuation,
	.line_output        = line_output,
	.line_other         = line_other,
	.processed_line     = processed_line,
	.finalize           = finalize,
	.cleanup            = cleanup,
};
