/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Parsing hooks for generating html
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "parse.h"
#include "wbuf.h"



struct gen_html_context {
	FILE *out;           /* output stream */
	unsigned int lines;  /* number of lines processed */
};


/* --------------------------------------------------------------------------
 *   private functions
 */

static void
print_escaped(const char *text, FILE *out)
{
	const char *p = text, *last = text;

	while (1) {
		if (strchr("<>\"'&\n", *p)) {
			if (p > last)
				fwrite(last, p - last, 1, out);

			if (!*p)
				break;

			switch (*p) {
			case '<': fputs("&lt;", out); break;
			case '>': fputs("&gt;", out); break;
			case '"': fputs("&quot;", out); break;
			case '\'': fputs("&#39;", out); break;
			case '&': fputs("&amp;", out); break;
			case '\n': break;
			default: assert(0); break;
			}

			last = p + 1;
		}

		p++;
	}
}



/* --------------------------------------------------------------------------
 *   parsing operations
 */

static void *
init(FILE *out)
{
	struct gen_html_context *ctx;

	ctx = malloc(sizeof(*ctx));
	if (!ctx)
		return NULL;

	ctx->out    = out;
	ctx->lines  = 0;

	fputs("<div class='kTest'>", out);

	return ctx;
}

static void
cleanup(void *priv)
{
	struct gen_html_context *ctx = (struct gen_html_context *)priv;

	fclose(ctx->out);
}

static int
line_command(void *priv,
             parse_state prev_state,
             char *line,
             char *data,
             size_t len)
{
	struct gen_html_context *ctx = (struct gen_html_context *)priv;
	FILE *out = ctx->out;

	switch (prev_state) {
	case STATE_OUTPUT:
		fputs("</pre></div>" /* kOut */
		      "</div>", out); /* kCmd */
		break;
	case STATE_COMMAND:
		fputs("</pre></div>", out); /* kCmd */
		break;
	case STATE_NONE:
		if (ctx->lines)
			fputs("</pre></div>", out); /* kTxt */
		break;
	default:
		assert(0);
	}
	fputs("<div class='kCmd'><pre>\n", out); /* kCmd */

	fputs("<span class='kCmP'>$ </span>", out);
	print_escaped(data, out);
	fputc('\n', out);

	return 0;
}

static int
line_continuation(void *priv,
                  parse_state prev_state,
                  char *line,
                  char *data,
                  size_t len)
{
	struct gen_html_context *ctx = (struct gen_html_context *)priv;
	FILE *out = ctx->out;

	fputs("<span class='kCmP'>&gt; </span>", out);
	print_escaped(data, out);
	fputc('\n', out);

	return 0;
}

static int
line_output(void *priv,
            parse_state prev_state,
            char *line,
            char *data,
            size_t len)
{
	struct gen_html_context *ctx = (struct gen_html_context *)priv;
	FILE *out = ctx->out;

	switch (prev_state) {
	case STATE_OUTPUT:
		break;
	case STATE_COMMAND:
		fputs("</pre>"
		      "<div class='kOut'><pre>\n", out);
		break;
	default:
		assert(0);
	}

	print_escaped(data, out);
	fputc('\n', out);

	return 0;
}

static int
line_other(void *priv,
           parse_state prev_state,
           char *line,
           char *data,
           size_t len)
{
	struct gen_html_context *ctx = (struct gen_html_context *)priv;
	FILE *out = ctx->out;

	switch (prev_state) {
	case STATE_OUTPUT:
		fputs("</pre></div>" /* kOut */
		      "</div>" /* kCmd */
		      "<div class='kTxt'><pre>\n", out);
		break;
	case STATE_COMMAND:
		fputs("</pre></div>" /* kCmd */
		      "<div class='kTxt'><pre>\n", out);
		break;
	case STATE_NONE:
		if (!ctx->lines)
			fputs("<div class='kTxt'><pre>\n", out);
		break;
	default:
		assert(0);
	}

	print_escaped(data, out);
	fputc('\n', out);

	return 0;
}

static void
processed_line(void *priv)
{
	struct gen_html_context *ctx = (struct gen_html_context *)priv;

	ctx->lines++;
}

static int
finalize(void *priv, parse_state prev_state)
{
	struct gen_html_context *ctx = (struct gen_html_context *)priv;
	FILE *out = ctx->out;

	switch (prev_state) {
	case STATE_OUTPUT:
		fputs("</pre></div>" /* kOut */
		      "</div>", out); /* kCmd */
		break;
	case STATE_COMMAND:
		fputs("</pre></div>", out); /* kCmd */
		break;
	case STATE_NONE:
		if (ctx->lines)
			fputs("</pre></div>", out); /* kTxt */
		break;
	default:
		assert(0);
	}

	fputs("</div>", out); /* kTest */

	return 0;
}



/* --------------------------------------------------------------------------
 *   operations table
 */

const struct parse_ops parse_ops_gen_html = {
	.init               = init,
	.set_indent         = NULL,
	.line_command       = line_command,
	.line_continuation  = line_continuation,
	.line_output        = line_output,
	.line_other         = line_other,
	.processed_line     = processed_line,
	.finalize           = finalize,
	.cleanup            = cleanup,
};
