/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Parsing hooks for reStructuredText generation
 *
 * The resulting output is meant to be later transformed into html
 * using rst2html.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "parse.h"


struct gen_rest_context {
	FILE *out;  /* output stream */
};

/* prefix to all raw html parts - allows the indentation to be preserved
 * no matter what the content of those parts actually is */
#define RAW_PREFIX  "  <!-- raw -->"


/* --------------------------------------------------------------------------
 *   private functions
 */

static void
print_escaped(const char *text, FILE *out)
{
	const char *p = text, *last = text;

	while (1) {
		if (strchr("<>\"'&\n", *p)) {
			if (p > last)
				fwrite(last, p - last, 1, out);

			if (!*p)
				break;

			switch (*p) {
			case '<': fputs("&lt;", out); break;
			case '>': fputs("&gt;", out); break;
			case '"': fputs("&quot;", out); break;
			case '\'': fputs("&#39;", out); break;
			case '&': fputs("&amp;", out); break;
			case '\n': break;
			default: assert(0); break;
			}

			last = p + 1;
		}

		p++;
	}
}



/* --------------------------------------------------------------------------
 *   parsing operations
 */

static void *
init(FILE *out)
{
	struct gen_rest_context *ctx;

	ctx = malloc(sizeof(*ctx));
	if (!ctx)
		return NULL;

	ctx->out = out;

	return ctx;
}

static void
cleanup(void *priv)
{
	struct gen_rest_context *ctx = (struct gen_rest_context *)priv;

	fclose(ctx->out);
}

static int
line_command(void *priv,
             parse_state prev_state,
             char *line,
             char *data,
             size_t len)
{
	struct gen_rest_context *ctx = (struct gen_rest_context *)priv;
	FILE *out = ctx->out;

	switch (prev_state) {
	case STATE_OUTPUT:
		fputs(RAW_PREFIX
		      "</pre></div>" /* kOut */
		      "</div>", out); /* kCmd */
		break;
	case STATE_COMMAND:
		fputs(RAW_PREFIX
		      "</pre></div>", out); /* kCmd */
		break;
	case STATE_NONE:
		fputs("\n.. raw:: html\n\n", out);
		break;
	default:
		assert(0);
	}
	fputs(RAW_PREFIX
	      "<div class='kCmd'><pre>\n", out); /* kCmd */

	fputs(RAW_PREFIX
	      "<span class='kCmP'>$ </span>", out);
	print_escaped(data, out);
	fputc('\n', out);

	return 0;
}

static int
line_continuation(void *priv,
                  parse_state prev_state,
                  char *line,
                  char *data,
                  size_t len)
{
	struct gen_rest_context *ctx = (struct gen_rest_context *)priv;
	FILE *out = ctx->out;

	fputs(RAW_PREFIX
	      "<span class='kCmP'>&gt; </span>", out);
	print_escaped(data, out);
	fputc('\n', out);

	return 0;
}

static int
line_output(void *priv,
            parse_state prev_state,
            char *line,
            char *data,
            size_t len)
{
	struct gen_rest_context *ctx = (struct gen_rest_context *)priv;
	FILE *out = ctx->out;

	switch (prev_state) {
	case STATE_OUTPUT:
		break;
	case STATE_COMMAND:
		fputs(RAW_PREFIX
		      "</pre>"
		      "<div class='kOut'><pre>\n", out);
		break;
	default:
		assert(0);
		return -1;
	}

	fputs(RAW_PREFIX, out);
	print_escaped(data, out);
	fputc('\n', out);

	return 0;
}

static int
line_other(void *priv,
           parse_state prev_state,
           char *line,
           char *data,
           size_t len)
{
	struct gen_rest_context *ctx = (struct gen_rest_context *)priv;
	FILE *out = ctx->out;

	switch (prev_state) {
	case STATE_OUTPUT:
		fputs(RAW_PREFIX
		      "</pre></div>" /* kOut */
		      "</div>\n\n", out); /* kCmd */
		break;
	case STATE_COMMAND:
		fputs(RAW_PREFIX
		      "</pre></div>\n\n", out); /* kCmd */
		break;
	case STATE_NONE:
		break;
	default:
		assert(0);
	}

	print_escaped(data, out);
	fputc('\n', out);

	return 0;
}

static int
finalize(void *priv, parse_state prev_state)
{
	struct gen_rest_context *ctx = (struct gen_rest_context *)priv;
	FILE *out = ctx->out;

	switch (prev_state) {
	case STATE_OUTPUT:
		fputs(RAW_PREFIX
		      "</pre></div>" /* kOut */
		      "</div>", out); /* kCmd */
		break;
	case STATE_COMMAND:
		fputs(RAW_PREFIX
		      "</pre></div>", out); /* kCmd */
		break;
	case STATE_NONE:
		break;
	default:
		assert(0);
	}

	fputs("\n\n", out);

	return 0;
}



/* --------------------------------------------------------------------------
 *   operations table
 */

const struct parse_ops parse_ops_gen_rest = {
	.init               = init,
	.set_indent         = NULL,
	.line_command       = line_command,
	.line_continuation  = line_continuation,
	.line_output        = line_output,
	.line_other         = line_other,
	.processed_line     = NULL,
	.finalize           = finalize,
	.cleanup            = cleanup,
};
