/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Core parsing of test files
 * Further processing is done by hooks provided by generators
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>

#include "parse.h"


struct parse_context {
	parse_state state;            /* current parsing state */
	parse_state prev_state;       /* previous parsing state */

	const struct parse_ops *ops;  /* operations performed while parsing */
	void *priv;                   /* private operations data */

	unsigned int indent;          /* indentation size */

	char *lbuf;                   /* line buffer */
	size_t lbuf_len;              /* line buffer length */

	FILE *in;                     /* input stream */
	FILE *out;                    /* output stream */
};



/* --------------------------------------------------------------------------
 *   private functions
 */

static bool
get_indent(struct parse_context *c, const char *line)
{
	const char *start, *p;

	start = strchr(line, '$');
	if (!start || start == line)
		return false;
	for (p = line; p != start; p++)
		if (*p != ' ')
			return false;

	c->indent = start - line;

	if (c->ops->set_indent)
		c->ops->set_indent(c->priv, c->indent);

	return true;
}

static int
process_line(struct parse_context *c, char *line, ssize_t len)
{
	const struct parse_ops *ops = c->ops;
	char *p, *start, *data;
	line_op op;

	c->prev_state = c->state;
	assert(c->prev_state != STATE_END);

	if (!c->indent && !get_indent(c, line))
		return ops->line_other ? ops->line_other(c->priv, STATE_NONE,
		                                         line, line, len)
		                       : 0;

	start = line + c->indent;
	for (p = line; *p && p != start; p++)
		if (*p != ' ')
			break;

	if (p < start) {
		/* no special indentation */
		c->state = STATE_NONE;
		op = ops->line_other;
		data = line;
	} else if (*start == '$' && start[1] == ' ') {
		/* command prefix */
		c->state = STATE_COMMAND;
		op = ops->line_command;
		data = start + 2;
	} else if (*start++ == '>' && (*start == ' ' || *start == '\n')) {
		/* command continuation prefix */
		if (c->prev_state == STATE_COMMAND) {
			c->state = STATE_COMMAND;
			op = ops->line_continuation;
			if (*start != '\n')
				++start;
			data = start;
		} else {
			c->state = STATE_NONE;
			op = ops->line_other;
			data = line;
		}
	} else {
		/* output prefix */
		if (c->prev_state == STATE_NONE) {
			c->state = STATE_NONE;
			op = ops->line_other;
			data = line;
		} else {
			c->state = STATE_OUTPUT;
			op = ops->line_output;
			data = start - 1;
		}
	}

	return op ? op(c->priv, c->prev_state, line, data, len - (data - line))
	          : 0;
}



/* --------------------------------------------------------------------------
 *   public functions
 */

struct parse_context *
parse_init(FILE *in, FILE *out, const struct parse_ops *ops)
{
	struct parse_context *c;
	void *priv;

	if ( !(c    = calloc(1, sizeof(*c))) ||
	     !(priv = ops->init(out)) )
	{
		fprintf(stderr, "Out of memory!\n");
		free(c);
		return NULL;
	}

	c->state       = STATE_NONE;
	c->ops         = ops;
	c->priv        = priv;
	c->indent      = 0;
	c->lbuf        = NULL;
	c->lbuf_len    = 0;
	c->in          = in;
	c->out         = out;

	return c;
}

/* returns the number of line processed (0 indicates end of input)
 * or -1 in case of error
 * 
 * memory and input stream both implicitly released when processing is over
 */
int
parse_process_line(struct parse_context *c)
{
	int res = 0;
	ssize_t len;

	len = getline(&c->lbuf, &c->lbuf_len, c->in);
	if (len != -1) {
		if (process_line(c, c->lbuf, len) == -1)
			goto out;

		if (c->ops->processed_line)
			c->ops->processed_line(c->priv);

		return 1; /* one line processed */
	}
	/* end of input */

	if (c->ops->finalize)
		res = c->ops->finalize(c->priv, c->state);

out:
	fclose(c->in);
	c->ops->cleanup(c->priv);
	free(c->lbuf);
	free(c);

	return res;
}

int
parse_process_stream(FILE *in, FILE *out, const struct parse_ops *ops)
{
	struct parse_context *c;
	int res;

	c = parse_init(in, out, ops);
	if (!c)
		return -1;

	while ( (res = parse_process_line(c)) > 0)
		;

	return res;
}
