/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Core parsing of test files
 * Further processing is done by hooks provided by generators
 */

#ifndef _KUULTO_PARSE_H
#define _KUULTO_PARSE_H

/**
 * Parsing states
 */
typedef enum {
	STATE_COMMAND,  /* within a command or command continuation */
	STATE_OUTPUT,   /* within command output */
	STATE_NONE,     /* neither command nor output */
	STATE_END       /* last line of input has been processed */
} parse_state;


typedef int (*line_op)(void *priv,
                       parse_state prev_state,
                       char *line,
                       char *data,
                       size_t len);

/**
 * Operations performed while parsing
 */
struct parse_ops {
	/* mandatory hooks */
	void *(*init)(FILE *out);
	void (*cleanup)(void *priv);

	/* optional hooks */
	line_op line_command;
	line_op line_continuation;
	line_op line_output;
	line_op line_other;
	void (*set_indent)(void *priv, unsigned int width);
	void (*processed_line)(void *priv);
	int (*finalize)(void *priv, parse_state prev_state);
};


struct parse_context;

/**
 * Initialize commands generation
 *
 * \param[in] in   Input stream: test file
 * \param[in] out  Output stream
 * \param[in] ops  Parsing operations
 *
 * \return         A handle suitable for use with parse_process_input()
 *                 or NULL in case of error
 */
struct parse_context *
parse_init(FILE *in, FILE *out, const struct parse_ops *ops);

/**
 * Process a line of test input
 *
 * Memory and I/O streams implicitly released when processing is over.
 *
 * \param[in] c  Handle obtained with parse_init()
 *
 * \return       The number of line processed (0 indicates end of input)
 *               or -1 in case of error
 */
int parse_process_line(struct parse_context *c);

/**
 * Process a whole test input stream
 *
 * Memory and I/O streams implicitly released when processing is over.
 *
 * \param[in] in   Input stream: test file
 * \param[in] out  Output stream
 * \param[in] ops  Parsing operations
 *
 * \return         0 if processing went ok, -1 otherwise
 *
 * this wraps parse_init() and parse_process_input() into a single
 * function for convenience
 */
int parse_process_stream(FILE *in, FILE *out, const struct parse_ops *ops);

#endif /* _KUULTO_PARSE_H */
