/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "parse.h"
#include "fdutils.h"
#include "shell.h"
#include "merge.h"

extern const struct parse_ops parse_ops_gen_cmds;



/* --------------------------------------------------------------------------
 *   private functions
 */

static int
run_test_initialized(int shfd_in,
                     int shfd_out,
                     struct parse_context *ctx_gen,
                     struct merge_context *ctx_merge)
{
	bool merge_initialized = false;
	fd_set rfds, wfds;
	int rc, nfds;

	nfds = 1 + ((shfd_in > shfd_out) ? shfd_in : shfd_out);

	while (1) {
		FD_ZERO(&rfds);
		FD_SET(shfd_out, &rfds);

		FD_ZERO(&wfds);
		if (shfd_in != -1)
			FD_SET(shfd_in, &wfds);

		while ((rc = select(nfds, &rfds, &wfds, NULL, NULL)) == -1)
			if (errno != EINTR) {
				perror("select");
				return -1;
			}

		if (FD_ISSET(shfd_out, &rfds)) {
			if (!merge_initialized) {
				if (merge_init(ctx_merge) == -1)
					return -1;
				merge_initialized = true;
			}

			/* merge shell output with test stream */
			rc = merge_process_result_line(ctx_merge);
			if (rc <= 0)
				return rc;
		}

		if (shfd_in != -1 && FD_ISSET(shfd_in, &wfds))
			/* produce commands for the shell */
			if (parse_process_line(ctx_gen) <= 0) {
				shfd_in = -1;
				nfds = 1 + shfd_out;
			}
	}
}



/* --------------------------------------------------------------------------
 *   public functions
 */

int
run_test(const char *path, FILE *out, unsigned int merge_flags, int *rc)
{
	FILE *parse_test_in, *parse_out, *merge_test_in, *merge_res_in;
	int shfd_in, shfd_out, res, status;
	struct merge_context *ctx_merge;
	struct parse_context *ctx_gen;
	const char *name, *p;
	pid_t shell_pid;

	if (!(parse_test_in = fopen_cloexec(path, "r")) ||
	    !(merge_test_in = fopen_cloexec(path, "r")))
		return -1;

	/* -1 to request new file descriptors for shell I/O */
	shfd_in = -1;
	shfd_out = -1;

	p = strrchr(path, '/');
	name = p ? p + 1 : path;

	shell_pid = start_shell(&shfd_in, &shfd_out, name);
	if (shell_pid == -1)
		return -1;

	parse_out = fdopen(shfd_in, "w");
	if (!parse_out) {
		perror("fdopen parse to shell");
		return -1;
	}
	ctx_gen = parse_init(parse_test_in, parse_out, &parse_ops_gen_cmds);
	if (!ctx_gen)
		return -1;

	merge_res_in = fdopen(shfd_out, "r");
	if (!merge_res_in) {
		perror("fdopen shell to merge");
		return -1;
	}

	ctx_merge = merge_alloc(merge_test_in,
	                        merge_res_in,
	                        out,
	                        merge_flags);
	if (!ctx_merge)
		return -1;

	res = run_test_initialized(shfd_in, shfd_out, ctx_gen, ctx_merge);

	while (waitpid(shell_pid, &status, 0) == -1)
		if (errno != EINTR) {
			perror("waitpid");
			return -1;
		}
	if (WIFEXITED(status)) {
		*rc = WEXITSTATUS(status);
	} else {
		fprintf(stderr, "Warning: Shell exited abnormally\n");
		return -1;
	}

	return res;
}
