/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "../parse.h"


extern const struct parse_ops parse_ops_gen_cmds;
extern const struct parse_ops parse_ops_gen_ansi;
extern const struct parse_ops parse_ops_gen_html;
extern const struct parse_ops parse_ops_gen_rest;
extern const struct parse_ops parse_ops_gen_script;

struct gen_type {
	const char * const name;      /**< Generator name */
	const struct parse_ops *ops;  /**< Associated parsing operations */
	const char * const *desc;     /**< Description */
};

#define TYPE_DESC(...)  (const char * const []) { __VA_ARGS__, NULL }
static const struct gen_type gen_types[] = {
	{
	  "cmds",
	  &parse_ops_gen_cmds,
	  TYPE_DESC("Annotated script whose output is meant to be merged back",
	            "into a test script later on using kuulto-merge. The",
	            "comparison between the original test and the merged test",
	            "will determine whether it passed or failed." )
	},
	{
	  "html",
	  &parse_ops_gen_html,
	  TYPE_DESC("Fragment of html usable in a web page.")
	},
	{
	  "rest",
	  &parse_ops_gen_rest,
	  TYPE_DESC("ReStructuredText format for html processing.")
	},
	{
	  "script",
	  &parse_ops_gen_script,
	  TYPE_DESC("Shell script with just commands and comments.")
	},
	{
	  "ansi",
	  &parse_ops_gen_ansi,
	  TYPE_DESC("ANSI terminal colors.")
	},
};

static const char * const usage_string =
	"This programs reads a test file and generates different kinds of\n"
	"outputs from it. If the test filename has been omitted or is '-',\n"
	"the test is read from the program's standard input.\n"
	"\n"
	"Options:\n"
	" -h         show this help text\n"
	" -t <type>  output type (see below, '%s' if unspecified)\n"
	"\n"
	"Output types:\n";


/* --------------------------------------------------------------------------
 *   private functions
 */

static void
__attribute__((noreturn))
usage(const char *argv0, const char *err)
{
	int i, j, rc = (err && *err) ? 1 : 0;

	fprintf(stderr, "\nUsage: %s [ <options> ] [ <test file> ]\n\n", argv0);
	fprintf(stderr, usage_string, (*gen_types).name);

	for (i = 0; i < sizeof(gen_types)/sizeof(*gen_types); i++) {
		const struct gen_type *t = &gen_types[i];
		for (j = 0; t->desc[j]; j++)
			fprintf(stderr, " %-11s%s\n", j ? "" : t->name,
			                              t->desc[j]);
	}

	if (rc != 0)
		fprintf(stderr, "\nError: %s\n", err);

	exit(rc);
}

static const struct parse_ops *
gen_ops_lookup(const char *name)
{
	int i;

	for (i = 0; i < sizeof(gen_types)/sizeof(*gen_types); i++)
		if (!strcmp(gen_types[i].name, name))
			return gen_types[i].ops;
	return NULL;
}



/* --------------------------------------------------------------------------
 *   public functions
 */

int
main(int argc, char **argv)
{
	FILE *in, *out = stdout;
	const struct parse_ops *ops = (*gen_types).ops;
	int opt, nargs;

	while ((opt = getopt(argc, argv, "ht:")) != -1)
		switch (opt) {
		case 'h':
			usage(argv[0], "");
		case 't':
			ops = gen_ops_lookup(optarg);
			if (!ops)
				usage(argv[0], "Invalid generator");
			break;
		default:
			usage(argv[0], "Invalid option");
		}
	nargs = argc - optind;

	if (nargs > 1) {
		usage(argv[0], "Too many arguments");
		return 1;
	}

	if (nargs == 1) {
		if (!strcmp(argv[optind], "-")) {
			in = stdin;
		} else {
			in = fopen(argv[optind], "r");
			if (!in) {
				perror("fopen");
				return 1;
			}
		}
	} else if (isatty(STDIN_FILENO)) {
		usage(argv[0], "No input specified");
		return 1;
	} else {
		in = stdin;
	}

	if (parse_process_stream(in, out, ops) == -1)
		return 1;

	return 0;
}
