/*
 * Copyright 2011 Cedric Duval <cedricduval@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2.
 *
 * See the accompanying COPYING file for more details.
 */

/**
 * \file
 * Write buffer for concatenating strings before printing them out
 *
 * Just a thin convenience wrapper around memory streams
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "wbuf.h"


struct wbuf {
	FILE *fp;     /* file pointer to write to */

	bool dirty;   /* is there unprinted text since last flush? */

	char *data;   /* data in memory stream */
	size_t len;   /* string length */
	off_t start;  /* offset to start of current set of lines */
};


/* --------------------------------------------------------------------------
 *   private functions
 */

static int
open_wbuf(struct wbuf *w)
{
	assert(!w->fp);

	w->fp = open_memstream(&w->data, &w->len);
	if (!w->fp)
	{
		perror("open_memstream");
		return -1;
	}
	w->start = 0;

	return 0;
}

static void
close_wbuf(struct wbuf *w)
{
	if (!w->fp)
		return;

	fclose(w->fp);
	w->fp = NULL;
	free(w->data);
	w->data = NULL;
}



/* --------------------------------------------------------------------------
 *   public functions
 */

struct wbuf *
wbuf_init()
{
	struct wbuf *w;

	w = malloc(sizeof(*w));
	if (!w) {
		fprintf(stderr, "Out of memory!\n");
		return NULL;
	}

	w->fp     = NULL;
	w->dirty  = false;
	w->data   = NULL;
	w->len    = 0;
	w->start  = 0;

	return w;
}

int
wbuf_yield(struct wbuf *w, int (*fn)(void *priv, char *data), void *priv)
{
	if (!w->fp || !w->dirty)
		return 0;

	/* process previously appended text(s) */
	if (fflush(w->fp) == EOF) {
		fprintf(stderr, "wbuf_print: %s", strerror(errno));
		return -1;
	}
	if ((*fn)(priv, w->data + w->start) != 0)
		return -1;

	w->start = w->len;
	w->dirty = false;

	/* a memory stream is typically given an initial size of BUFSIZ,
	 * which is comfortable enough for accumulating a fair number of
	 * lines; we only prevent it from growing too large */
	if (w->start > ((BUFSIZ * 95) / 100)) {
		fclose(w->fp);
		w->fp = NULL;
		free(w->data);
		open_wbuf(w);
	}

	return 0;
}

bool
wbuf_empty(struct wbuf *w)
{
	return !w->fp || !w->dirty;
}

int
wbuf_append(struct wbuf *w, const char *text)
{
	if (!w->fp)
		if (open_wbuf(w) == -1)
			return -1;

	assert(w->fp);

	if (fputs(text, w->fp) == EOF) {
		perror("fputs");
		return -1;
	}
	w->dirty = true;

	return 0;
}

void
wbuf_free(struct wbuf *w)
{
	close_wbuf(w);
	free(w);
}
